function make_BoundaryConditions(rhs::RHS, prms)
  type = query(String, prms, "type")

  xrange = extrema(rhs.mesh.x)
  broadcast_volume!(scaledcons2cons, rhs.equation, rhs.cache)

  lhs_bc, rhs_bc = @match type begin
    "periodic"          => (PeriodicBC(), PeriodicBC())
    "from initial data" => begin
      @unpack sD, sSr, stau = get_dynamic_variables(rhs.cache)
      @unpack D, Sr, tau, A, B, r, rho, vr, eps, p = get_static_variables(rhs.cache)
      lhs_state = (sD[1], sSr[1], stau[1], p[1], rho[1], vr[1], eps[1], A[1], B[1], r[1])
      rhs_state = (sD[end], sSr[end], stau[end], p[end], rho[end], vr[end], eps[end], A[end], B[end], r[end])
      (DirichletBC(lhs_state), DirichletBC(rhs_state))
    end
    _ => error("Boundary condition type '$type' not implemented!")
  end

  return BoundaryConditions(lhs_bc, rhs_bc, rhs.rsolver)
end


function make_BoundaryConditions_LDG(ldg_rsolver, prms)
  isnothing(ldg_rsolver) && return nothing
  type = query(String, prms, "type")
  lhs_bc, rhs_bc = if type == "periodic"
    PeriodicBC(), PeriodicBC()
  else
    OutflowBC(), OutflowBC()
  end
  return BoundaryConditions(lhs_bc, rhs_bc, ldg_rsolver)
end
